#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sched.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#define TDIFF(t0,t1) ((t1.tv_sec - t0.tv_sec)*1000000 + t1.tv_usec-t0.tv_usec)

void wrrd(int wrfd, int rdfd, int count, int readfirst)
{
    char c;
    if (readfirst) {read(rdfd, &c, 1); count--;}
    while (count--) {
	write(wrfd, &c, 1);
	read(rdfd, &c, 1);
    }
    if (readfirst) write(wrfd, &c, 1);

}

int main(int argc, char **argv)
{
    struct timeval tv1,tv2;
    int pfd[2][2];
    int t1, t2, few, gtod, count;
    struct sched_param param;

    if (argc != 2) exit(1);
    if (!(count = atoi(argv[1]))) exit(2);

    if (mlockall(MCL_CURRENT | MCL_FUTURE)) {
        fprintf(stderr, "%s: mlockall: %s\n", argv[0], strerror(errno));
        /* continue anyway */
    }
    if (!getenv("BENCH_NOTRT")) {
	param.sched_priority = 1;
	if (sched_setscheduler(0, SCHED_FIFO, &param)<0) {
	    fprintf(stderr,"%s: setscheduler: %s\n", argv[0], strerror(errno));
	    /* continue anyway */
	}
    }
    pipe(pfd[0]);
    pipe(pfd[1]);

    gettimeofday(&tv1, NULL); /* first one for the cache */
    gettimeofday(&tv1, NULL);
    gettimeofday(&tv2, NULL);
    gtod = TDIFF(tv1,tv2);
    printf("gettimeofday takes %i usecs\n", gtod);

    if (!fork()) exit(0); else wait(&t1); /* first one for the cache */
    gettimeofday(&tv1, NULL);
    if (!fork()) exit(0); else wait(&t1);
    gettimeofday(&tv2, NULL);
    few = TDIFF(tv1,tv2);
    printf("fork/exit/wait takes %i usecs\n", few);

    if (!getenv("CSWITCH_NOTALONE")) {
	gettimeofday(&tv1, NULL);
	wrrd(pfd[0][1],pfd[0][0], count, 0);
	gettimeofday(&tv2, NULL);
	t1 =  TDIFF(tv1,tv2) - gtod;
	printf("%i writes and reads take %i usecs (%i.%03i usec each)\n",count,
	       t1, t1/count, ((t1%count * 1000 +count/2) / count));
    }

    gettimeofday(&tv1, NULL);
    if (!fork()) {
	wrrd(pfd[1][1], pfd[0][0], count/2, 0);
	exit(0);
    }
    wrrd(pfd[0][1], pfd[1][0], count/2, 1);
    wait(&t2);
    gettimeofday(&tv2, NULL);
    t2 =  TDIFF(tv1,tv2) - gtod - few;
    printf("%i wr/rd and cswitches take %i usecs (%i.%03i usec each)\n", count,
	   t2, t2/count, ((t2%count * 1000 +count/2) / count));
    t2 -= t1;
    printf("%i cswitches take %i usecs (%i.%03i usec each)\n", count,
	   t2, t2/count, ((t2%count * 1000 +count/2) / count));
    exit(0);
}
